
def read_varlen(data):
    NEXTBYTE = 1
    value = 0
    while NEXTBYTE:
        chr = ord(bytearray([next(data)]))
        # is the hi-bit set?
        if not (chr & 0x80):
            # no next BYTE
            NEXTBYTE = 0
        # mask out the 8th bit
        chr = chr & 0x7f
        # shift last value up 7 bits
        value = value << 7
        # add new value
        value += chr
    return value


def write_varlen(value):
    chr1 = bytearray([value & 0x7F])
    value >>= 7
    if value:
        chr2 = bytearray([(value & 0x7F) | 0x80])
        value >>= 7
        if value:
            chr3 = bytearray([(value & 0x7F) | 0x80])
            value >>= 7
            if value:
                chr4 = bytearray([(value & 0x7F) | 0x80])
                res = chr4 + chr3 + chr2 + chr1
            else:
                res = chr3 + chr2 + chr1
        else:
            res = chr2 + chr1
    else:
        res = chr1
    return res



import os
import sys

__version__ = "0.0.1.dev7"


def find_beatsearch_project_dir(max_levels: int = 5) -> str:
    cd = os.path.dirname(os.path.realpath(__file__))

    def check_directory(directory):
        return all(os.path.isfile(os.path.join(directory, fname)) for fname in [
            "build_dependencies.py", "setup.py"
        ])

    for level in range(max_levels):
        if check_directory(cd):
            return cd
        next_dir = os.path.dirname(cd)
        if next_dir == cd:
            break
        cd = os.path.dirname(cd)

    return ""


BEATSEARCH_PROJECT_DIR = find_beatsearch_project_dir()
if BEATSEARCH_PROJECT_DIR:
    sys.path.append(BEATSEARCH_PROJECT_DIR)
    from build_dependencies import Python3Midi, get_third_party_dir

    python3_midi = Python3Midi(get_third_party_dir(BEATSEARCH_PROJECT_DIR))

    if not python3_midi.has_build():
        print("Python3Midi hasn't built yet, please run build_dependencies.py")
        sys.exit(-1)

    try:
        import midi
    except ModuleNotFoundError:
        sys.path.append(python3_midi.lib_dir)



# coding=utf-8
import os
import sys
import collections
import typing as tp
from time import time
from abc import ABCMeta, abstractmethod
from inspect import isclass
from functools import wraps, reduce
from matplotlib.colors import to_rgb, rgb_to_hsv, hsv_to_rgb, to_hex


def merge_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


def format_timespan(seconds):
    """
    Returns the given timespan as a string in hours:minutes:seconds.

    :param seconds: timespan in seconds
    :return: formatted string
    """

    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return "%d:%02d:%02d" % (h, m, s)


def print_progress_bar(iteration, total, prefix="", suffix="", decimals=1, length=50, fill='█', starting_time=None):
    """
    Call in a loop to create terminal progress bar

    :param iteration: current iteration
    :param total: total iterations
    :param prefix: prefix string
    :param suffix: suffix string
    :param decimals: number of decimals in percent progress
    :param length: character length of bar
    :param fill: bar fill character
    :param starting_time: starting time, for computing the time remaining for completion
    :return: terminal progress bar
    """

    progress = iteration / float(total)
    percent = ("{0:." + str(decimals) + "f}").format(100 * progress)
    filled_length = int(length * iteration // total)
    bar = fill * filled_length + '-' * (length - filled_length)

    if starting_time is not None:
        elapsed_time = time() - starting_time
        if iteration < total:
            try:
                remaining = format_timespan((elapsed_time / progress) - elapsed_time)
            except ZeroDivisionError:
                remaining = "unknown"
            time_info = "Remaining: %s " % remaining
        else:
            time_info = "Elapsed: %s" % format_timespan(elapsed_time)
    else:
        time_info = ""

    sys.stdout.write('\r%s |%s| %s%% %s %s' % (prefix, bar, percent, suffix, time_info))
    sys.stdout.flush()

    if iteration == total:
        print("")


def friendly_named_class(name):
    # noinspection PyPep8Naming
    def _friendly_named_class_decorator(Cls):
        setattr(Cls, '__friendly_name__', name)
        return Cls

    return _friendly_named_class_decorator


def err_print(*args, **kwargs):
    """
    Print function wrapper for printing to stderr.
    """

    print(*args, file=sys.stderr, **kwargs)


def make_dir_if_not_exist(directory: str) -> str:
    """
    Creates a directory if it doesn't exist yet.

    :param directory: path
    :return: path
    """

    if not os.path.isdir(directory):
        os.mkdir(directory)
    return directory


def head_trail_iter(iterable):
    """
    Returns an iterator which yields a tuple (is_first, is_last, item) on each iteration.

    :param iterable: iterable with length
    :return: iterator returning (is_first, is_last, item)
    """

    last_i = len(iterable) - 1
    for i, item in enumerate(iterable):
        first = i == 0
        last = i == last_i
        yield first, last, item


def get_beatsearch_dir(mkdir=True):
    """Returns the beatsearch directory
    Returns the beatsearch directory. This directory is used to store *.ini files and output various beatsearch output
    data.

    :param mkdir: when True, this function will create the beatsearch directory in case it does not exist
    :return: beatsearch directory path
    """

    home = os.path.expanduser("~")
    beatsearch_dir = os.path.join(home, "beatsearch")
    if not os.path.isdir(beatsearch_dir) and mkdir:
        os.mkdir(beatsearch_dir)
    return beatsearch_dir


def get_default_beatsearch_rhythms_fpath(mkdir=True):
    """Returns the default path to the rhythms pickle file

    Returns the default path to the rhythms pickle file. This path will be ${BSHome}/data/rhythms.pkl, where BSHome
    is the beatsearch directory returned by get_beatsearch_dir.

    :return: default path to the rhythms pickle file as a string
    """

    beatsearch_dir = get_beatsearch_dir(mkdir)
    return os.path.join(beatsearch_dir, "rhythms.pkl")


def no_callback(*_, **__):
    return None


def type_check_and_instantiate_if_necessary(thing, thing_base_type, allow_none=True, *args, **kwargs):
    """
    Instantiates the given 'thing' with the given args and kwargs and returns it. If the given 'thing' is already an
    instance of the given 'thing_base_type' it will return 'thing'.

    :param thing: either an instance or a (sub)-class of 'thing_base_type'
    :param thing_base_type: a class
    :param allow_none: when True, this function will allow None things (it will return None if 'thing' is None)
    :param args: if an initialization is needed, this positional arguments will be passed to the constructor
    :param kwargs: if an initialization is needed, this named arguments will be passed to the constructor
    :return: an instance of 'thing_base_type' (might return None if 'allow_none' is True)
    """
    # type: (tp.Any, tp.Type, bool) -> tp.Any

    if not isclass(thing_base_type):
        raise TypeError("Expected a class but got \"%s\"" % str(thing_base_type))

    if (allow_none and thing is None) or isinstance(thing, thing_base_type):
        return thing

    if not isclass(thing) or not issubclass(thing, thing_base_type):
        raise TypeError("Expected either a \"%s\" (sub)-class or "
                        "an instance but got \"%s\"" % (thing_base_type, str(thing)))

    return thing(*args, **kwargs)


def eat_args(func):
    """
    Adds ignored positional and named arguments to the given function's signature.

    :param func: function to add the fake arguments to
    :return: wrapped function
    """

    @wraps(func)
    def wrapper(*_, **__):
        return func()
    return wrapper


def color_variant(color, brightness_offset=0):
    """
    Takes a color and produces a lighter or darker variant.

    :param color: color to convert (either as a rgb array or as a hexadecimal string)
    :param brightness_offset: brightness offset between -1 and 1
    :return: new color (either hexadecimal or rgb depending on whether the input is a hexadecimal color)
    """

    try:
        is_hexadecimal = color.startswith("#")
    except TypeError:
        is_hexadecimal = False

    rgb = to_rgb(color)
    hsv = rgb_to_hsv(rgb)

    curr_brightness = hsv[2]
    new_brightness = max(-1.0, min(curr_brightness + brightness_offset, 1.0))  # clamp to [-1.0, 1.0]

    new_hsv = (hsv[0], hsv[1], new_brightness)
    new_rgb = hsv_to_rgb(new_hsv)

    if is_hexadecimal:
        return to_hex(new_rgb)

    return new_rgb


def get_midi_files_in_directory(directory):
    """Iterates over the midi files in the given directory

    Recursively iterates oer the midi files in the given directory yielding the full MIDI-file paths.

    :param directory: directory to iterate over
    :return: generator yielding paths to the MIDI files
    """

    for directory, subdirectories, files in os.walk(directory):
        for f_name in files:
            if os.path.splitext(f_name)[1] != ".mid":
                continue
            yield os.path.join(directory, directory, f_name)


class TupleView(collections.Sequence):
    @staticmethod
    def range_view(the_tuple, from_index, to_index):
        """Creates a range tuple view

        :param the_tuple: the tuple to create the view of
        :param from_index: the starting index
        :param to_index: the ending index (excluding)
        :return: TupleView
        """

        assert to_index >= from_index
        indices = range(from_index, to_index) if to_index > from_index else tuple()
        return TupleView(the_tuple, indices)

    def __init__(self, the_tuple, indices):
        """Creates a view of the given tuple

        :param the_tuple: the tuple to create the view of
        :param indices: the indices of the tuple
        """

        indices = tuple(indices)

        for ix in indices:
            try:
                the_tuple[ix]
            except IndexError:
                raise ValueError("invalid index: %i" % ix)

        self._the_tuple = the_tuple
        self._indices = indices

    def __getitem__(self, index):
        actual_index = self._indices[index]
        return self._the_tuple[actual_index]

    def __len__(self):
        return len(self._indices)


def most_common_element(sequence: tp.Sequence[tp.Any]) -> tp.Any:
    """Returns the most common element in the given sequence

    :param sequence: sequence
    :return: None
    """

    return max(set(sequence), key=sequence.count)


def sequence_product(iterable):
    """Returns the product of the given iterable

    Returns 0 if reduce() raise a TypeError.

    :param iterable: iterable to return the product of
    :return: product of given iterable
    """

    try:
        return reduce((lambda x, y: x * y), iterable)
    except TypeError:
        return 0


def minimize_term_count(value: int, terms: tp.Sequence[int], assume_sorted: bool = False):
    """This function minimizes the number of terms necessary to sum up a value, given a set of terms

    This function returns terms that eventually sum up to the given value. It is only allowed to use terms given in the
    "terms" parameter or 1. For example, if given value 17 and terms [3, 5, 7], this method will yield [7, 7, 5, 1].

    :param value: the value that the terms should sum up to as an integer
    :param terms: sequence containing the terms to choose from
    :param assume_sorted: set this to true if the given terms are sorted (in ascending order) for better performance
    :return: a generator yielding maximized terms that sum up to the given value
    """

    if not assume_sorted:
        terms = sorted(terms)
    i = 0
    while i < value:
        remaining = value - i
        u = next((u for u in reversed(terms) if u <= remaining), 1)
        n = remaining // u
        for j in range(n):
            yield u
        i += n * u


FileInfo = collections.namedtuple("FileInfo", ["path", "modified_time"])


def normalize_directory(directory):
    """Normalizes a directory path

    Normalizes the directory path, replaces the backslashes with forward slashes and removes the ending slash.

    :param directory: directory path to normalize
    :return: normalized directory path
    """

    directory = os.path.normpath(directory)
    directory = directory.replace("\\", "/")
    if directory.endswith("/"):
        directory = directory[:-1]
    return directory


class Quantizable(object, metaclass=ABCMeta):
    @property
    def quantize_enabled(self) -> bool:
        return self.is_quantize_enabled()

    @quantize_enabled.setter
    def quantize_enabled(self, quantize_enabled: bool):
        self.set_quantize_enabled(quantize_enabled)

    @abstractmethod
    def set_quantize_enabled(self, quantize_enabled: bool):
        raise NotImplementedError

    @abstractmethod
    def is_quantize_enabled(self) -> bool:
        raise NotImplementedError


__all__ = [
    'merge_dicts', 'format_timespan', 'print_progress_bar', 'friendly_named_class',
    'err_print', 'make_dir_if_not_exist', 'head_trail_iter', 'get_beatsearch_dir',
    'get_default_beatsearch_rhythms_fpath', 'no_callback', 'type_check_and_instantiate_if_necessary',
    'eat_args', 'color_variant', 'get_midi_files_in_directory', 'TupleView', 'most_common_element',
    'sequence_product', 'minimize_term_count', 'FileInfo', 'normalize_directory', 'Quantizable'
]


import enum
import types
import itertools
import numpy as np
import typing as tp
from fractions import Fraction
from collections import OrderedDict, defaultdict
from abc import ABCMeta, abstractmethod
from beatsearch.utils import Quantizable, minimize_term_count
from beatsearch.rhythm import Rhythm, MonophonicRhythm, PolyphonicRhythm, \
    Unit, UnitType, parse_unit_argument, convert_tick


######################################
# Feature extractor abstract classes #
######################################


class FeatureExtractor(object, metaclass=ABCMeta):
    @abstractmethod
    def process(self, obj) -> tp.Any:
        """Computes and returns a feature
        Computes a feature of the given object and returns it.

        :param obj: object of which to compute the feature
        :return: the feature value
        """

        raise NotImplementedError


class RhythmFeatureExtractor(FeatureExtractor, metaclass=ABCMeta):
    @abstractmethod
    def get_pre_processors(self):  # type: () -> tp.Tuple[RhythmFeatureExtractor, ...]
        """Returns the preprocessors of this rhythm feature extractor

        Preprocessors are also instances of RhythmFeatureExtractor. When calling process(), the preprocessors will
        process the given rhythm and their results will be passed down to __process__.

        :return: tuple with this rhythm feature extractor's preprocessors
        """

        raise NotImplementedError

    @abstractmethod
    def get_unit(self) -> tp.Union[Unit, None]:
        """Returns the musical unit of this preprocessor

        :return: musical unit of this preprocessor as a string
        """

        raise NotImplementedError

    @abstractmethod
    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        """Sets the musical unit of this preprocessor

        :return:
        """

        raise NotImplementedError

    def process(self, rhythm: Rhythm) -> tp.Any:
        """Computes and returns a rhythm feature

        :param rhythm: rhythm of which to compute the feature
        :return: the rhythm feature value
        """

        pre_processor_results = list(pre_processor.process(rhythm) for pre_processor in self.get_pre_processors())
        ret = self.__process__(rhythm, pre_processor_results)
        return tuple(ret) if isinstance(ret, types.GeneratorType) else ret

    @abstractmethod
    def __process__(self, rhythm: Rhythm, pre_processor_results: tp.List[tp.Any]):
        """Computes and returns a rhythm feature

        If this method is implemented as a generator, it will be evaluated by process() and returned as a tuple.

        :param rhythm:                 rhythm of which to compute the feature
        :param pre_processor_results:  results of the preprocessors, which process the given rhythm
                                       immediately before the call to this method

        :return: the rhythm feature value
        """

        raise NotImplementedError

    @staticmethod
    def init_pre_processors():
        """
        Override this method to add preprocessors to the feature extractor. Preprocessors are RhythmFeatureExtractorBase
        objects themselves and their results will be passed down to __process__ as the last arguments.

        :return: an iterable of RhythmFeatureExtractorBase instances
        """

        return []

    ##############
    # properties #
    ##############

    @property
    def pre_processors(self):  # type: () -> tp.Tuple[RhythmFeatureExtractor, ...]
        """See get_pre_processors. This is a read-only property."""
        return self.get_pre_processors()

    @property
    def unit(self) -> tp.Union[Unit, None]:
        """See get_unit and set_unit"""
        return self.get_unit()

    @unit.setter
    def unit(self, unit: tp.Optional[UnitType]):
        self.set_unit(unit)


class RhythmFeatureExtractorBase(RhythmFeatureExtractor, metaclass=ABCMeta):
    def __init__(self, unit: tp.Optional[UnitType] = Unit.EIGHTH):
        self._pre_processors = tuple(self.init_pre_processors())  # type: tp.Tuple[RhythmFeatureExtractorBase, ...]
        self.__tick_to_unit__ = None

        self._unit = None  # type: tp.Union[Unit, None]
        self.unit = unit  # calling setter, which also sets the unit of the preprocessors

    def get_pre_processors(self) -> tp.Tuple[RhythmFeatureExtractor, ...]:
        """Returns the preprocessors, whose result is passed to __process__.

        Preprocessors are also rhythm feature extractors. Their unit is linked to this extractor's unit.
        """

        return self._pre_processors

    def get_unit(self) -> tp.Union[Unit, None]:
        """Returns the time unit of this feature extractor"""
        return self._unit

    @parse_unit_argument
    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        """Sets the time unit of this feature extractor

        Sets the time unit of this rhythm feature extractor. This method will also set the unit of the preprocessors.

        :param unit: new unit
        :return: None
        """

        for pre_processor in self._pre_processors:
            pre_processor.set_unit(unit)

        self._unit = unit


class QuantizableRhythmFeatureExtractorMixin(RhythmFeatureExtractor, Quantizable, metaclass=ABCMeta):
    def __init__(self, quantize_enabled):
        self._quantize_enabled = None
        self.set_quantize_enabled(quantize_enabled)

    def set_quantize_enabled(self, quantize_enabled: bool):
        quantize_enabled = bool(quantize_enabled)
        for feature_extractor in self.pre_processors:
            if not isinstance(feature_extractor, Quantizable):
                continue
            feature_extractor.set_quantize_enabled(quantize_enabled)
        self._quantize_enabled = quantize_enabled

    def is_quantize_enabled(self) -> bool:
        return self._quantize_enabled


class MonophonicRhythmFeatureExtractor(RhythmFeatureExtractorBase, metaclass=ABCMeta):
    @abstractmethod
    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """Computes and returns a monophonic rhythm feature
        Computes a feature of the given monophonic rhythm and returns it.

        :param rhythm: monophonic rhythm of which to compute the feature
        :return: the monophonic rhythm feature value
        """

        raise NotImplementedError


class PolyphonicRhythmFeatureExtractor(RhythmFeatureExtractorBase, metaclass=ABCMeta):
    @abstractmethod
    def __process__(
            self,
            rhythm: PolyphonicRhythm,
            pre_processor_results: tp.List[tp.Any]
    ):
        """Computes and returns a polyphonic rhythm feature
        Computes a feature of the given polyphonic rhythm and returns it.

        :param rhythm: polyphonic rhythm of which to compute the feature
        :param unit: concrete time unit
        :return: the polyphonic rhythm feature value
        """

        raise NotImplementedError


####################################################
# Generic rhythm feature extractor implementations #
####################################################


#######################################################
# Monophonic rhythm feature extractor implementations #
#######################################################


class BinaryOnsetVector(MonophonicRhythmFeatureExtractor):
    def __process__(self, rhythm: MonophonicRhythm, _):
        """
        Returns the binary representation of the note onsets of the given rhythm where each step where a note onset
        happens is denoted with a 1; otherwise with a 0. The given resolution is the resolution in PPQ (pulses per
        quarter note) of the binary vector.

        :param rhythm: the monophonic rhythm from which to compute the binary onset vector
        :return: the binary representation of the note onsets in the given rhythm
        """

        Rhythm.Precondition.check_resolution(rhythm)

        unit = self.unit
        n_steps = rhythm.get_duration(unit, ceil=True)
        binary_string = [0] * n_steps

        if unit is None:
            get_onset_position = lambda t: t
        else:
            resolution = rhythm.get_resolution()
            get_onset_position = lambda t: unit.from_ticks(t, resolution, True)

        for onset in rhythm.get_onsets():
            onset_position = get_onset_position(onset.tick)
            try:
                binary_string[onset_position] = 1
            except IndexError:
                pass  # when quantization pushes note after end of rhythm

        return binary_string


class NoteVector(MonophonicRhythmFeatureExtractor, QuantizableRhythmFeatureExtractorMixin):
    REST = 0
    NOTE = 1

    def __init__(self, unit: UnitType = Unit.EIGHTH, ret_positions: bool = False):
        super().__init__(unit)
        self._ret_positions = None
        self.ret_positions = ret_positions  # call setter

    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        if unit is None:
            raise ValueError("NoteVector does not support tick-based computation")
        super().set_unit(unit)

    @staticmethod
    def init_pre_processors():
        yield BinaryOnsetVector()

    @property
    def ret_positions(self):
        return self._ret_positions

    @ret_positions.setter
    def ret_positions(self, ret_position):
        self._ret_positions = bool(ret_position)

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        # TODO Add documentation

        Rhythm.Precondition.check_time_signature(rhythm)
        time_sig = rhythm.get_time_signature()

        natural_duration_map = time_sig.get_natural_duration_map(self.unit)
        duration_pool = sorted(set(natural_duration_map))  # NOTE The rest of this method depends on this being sorted
        binary_vector = pre_processor_results[0]

        if self._ret_positions:
            get_yield_values = lambda event_type, event_duration, event_position: \
                (event_type, event_duration, event_position)
        else:
            get_yield_values = lambda event_type, event_duration, event_position: (event_type, event_duration)

        i, n = 0, len(binary_vector)

        while i < n:
            metrical_position = i % len(natural_duration_map)
            natural_duration = natural_duration_map[metrical_position]
            is_onset = binary_vector[i]

            # compute the duration till the next note trimmed at the natural duration
            max_duration = next((j for j in range(1, natural_duration) if binary_vector[i + j]), natural_duration)

            if is_onset:
                # get the maximum available duration that fits the max duration
                note_duration = next(d for d in reversed(duration_pool) if d <= max_duration)
                yield get_yield_values(self.NOTE, note_duration, i)

                i += note_duration
                continue

            rests = tuple(minimize_term_count(max_duration, duration_pool, assume_sorted=True))

            for rest_duration in rests:
                yield get_yield_values(self.REST, rest_duration, i)
                i += rest_duration


class IOIVector(MonophonicRhythmFeatureExtractor, QuantizableRhythmFeatureExtractorMixin):
    class Mode(enum.Enum):
        PRE_NOTE = enum.auto()
        POST_NOTE = enum.auto()

    def __init__(self, unit: tp.Optional[UnitType] = Unit.EIGHTH, mode: Mode = Mode.POST_NOTE, quantize_enabled=True):
        MonophonicRhythmFeatureExtractor.__init__(self, unit)
        QuantizableRhythmFeatureExtractorMixin.__init__(self, quantize_enabled)
        self._mode = None
        self.mode = mode  # call setter

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode: Mode):
        self._mode = mode

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Returns the time difference between the notes in the given rhythm. The elements of the vector will depend on
        this IOIVector extractor's mode property:

        PRE_NOTE
            Time difference between the current note and the previous note. The first note will return the time
            difference with the start of the rhythm.

            For example, given the Rumba Clave rhythm:
              X--X---X--X-X---
              0  3   4  3 2

        POST_NOTE
            Time difference between the current note and the next note. The last note will return the time difference
            with the end (duration) of the rhythm.

            For example, given the Rumba Clave rhythm:
              X--X---X--X-X---
              3  4   3  2 4

        :param rhythm: the monophonic rhythm from which to compute the inter-onset interval vector
        :return: inter-onset interval vector of the given rhythm
        """

        Rhythm.Precondition.check_resolution(rhythm)

        # TODO Implement both modes generically here

        if self.mode == self.Mode.PRE_NOTE:
            return self._pre_note_ioi(rhythm)
        elif self.mode == self.Mode.POST_NOTE:
            return self._post_note_ioi(rhythm)
        else:
            raise RuntimeError("Unknown mode: \"%s\"" % self.mode)

    def _pre_note_ioi(self, rhythm: MonophonicRhythm):
        quantize = self.is_quantize_enabled()
        resolution = rhythm.get_resolution()
        unit = self.get_unit()
        current_tick = 0
        intervals = []

        for onset in rhythm.get_onsets():
            delta_tick = onset.tick - current_tick
            intervals.append(convert_tick(delta_tick, resolution, unit, quantize))
            current_tick = onset.tick

        return intervals

    # TODO Add cyclic option to include the offset in the last onset's interval
    def _post_note_ioi(self, rhythm: MonophonicRhythm):
        quantize = self.is_quantize_enabled()
        resolution = rhythm.get_resolution()
        unit = self.get_unit()
        intervals = []
        onset_positions = itertools.chain((onset.tick for onset in rhythm.get_onsets()), [rhythm.duration_in_ticks])
        last_onset_tick = -1

        for onset_tick in onset_positions:
            if last_onset_tick < 0:
                last_onset_tick = onset_tick
                continue

            delta_in_ticks = onset_tick - last_onset_tick
            delta_in_units = convert_tick(delta_in_ticks, resolution, unit, quantize)
            intervals.append(delta_in_units)
            last_onset_tick = onset_tick

        return intervals


class IOIHistogram(MonophonicRhythmFeatureExtractor):
    def __init__(self, unit: tp.Optional[UnitType] = Unit.EIGHTH):
        super().__init__(unit)

    @staticmethod
    def init_pre_processors():
        return [IOIVector(mode=IOIVector.Mode.POST_NOTE, quantize_enabled=True)]

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Returns the number of occurrences of the inter-onset intervals of the notes of the given rhythm in ascending
        order. The inter onset intervals are computed in POST_NOTE mode.

        For example, given the Rumba Clave rhythm, with inter-onset vector [3, 4, 3, 2, 4]:
            (
                [1, 2, 2],  # occurrences
                [2, 3, 4]   # bins (interval durations)
            )


        :return: an (occurrences, bins) tuple
        """

        ioi_vector = pre_processor_results[0]
        histogram = np.histogram(ioi_vector, tuple(range(min(ioi_vector), max(ioi_vector) + 2)))
        occurrences = histogram[0].tolist()
        bins = histogram[1].tolist()[:-1]
        return occurrences, bins


class BinarySchillingerChain(MonophonicRhythmFeatureExtractor):
    def __init__(self, unit: tp.Optional[UnitType] = Unit.EIGHTH, values=(1, 0)):
        super().__init__(unit)
        self._values = None
        self.values = values  # calls setter

    @staticmethod
    def init_pre_processors():
        yield BinaryOnsetVector()

    @property
    def values(self):
        """The two values of the schillinger chain returned by process

        Binary vector to be used in the schillinger chain returned by process(). E.g. when set to ('a', 'b'), the
        schillinger chain returned by process() will consist of 'a' and 'b'.
        """

        return self._values

    @values.setter
    def values(self, values):
        values = iter(values)
        self._values = next(values), next(values)

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Returns the Schillinger notation of this rhythm where each onset is a change of a "binary note".

        For example, given the Rumba Clave rhythm and with values (0, 1):
          X--X---X--X-X---
          0001111000110000

        However, when given the values (1, 0), the schillinger chain will be the opposite:
          X--X---X--X-X---
          1110000111001111

        :return: Schillinger rhythm vector as a list
        """

        values = self._values
        chain = list(pre_processor_results[0])
        i, value_i = 0, 1

        while i < len(chain):
            if chain[i] == 1:
                value_i = 1 - value_i
            chain[i] = values[value_i]
            i += 1

        return chain


class ChronotonicChain(MonophonicRhythmFeatureExtractor):
    @staticmethod
    def init_pre_processors():
        yield BinaryOnsetVector()

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Returns the chronotonic chain representation of the given rhythm.

        For example, given the Rumba Clave rhythm:
          X--X---X--X-X---
          3334444333224444

        :return: the chronotonic chain as a list
        """

        chain = pre_processor_results[0]  # binary onset vector
        i, delta = 0, 0

        while i < len(chain):
            if chain[i] == 1:
                j = i + 1
                while j < len(chain) and chain[j] == 0:
                    j += 1
                delta = j - i
            chain[i] = delta
            i += 1

        return chain


class IOIDifferenceVector(MonophonicRhythmFeatureExtractor, QuantizableRhythmFeatureExtractorMixin):
    def __init__(self, unit: tp.Optional[UnitType] = Unit.EIGHTH, quantize_enabled=True, cyclic=True):
        MonophonicRhythmFeatureExtractor.__init__(self, unit)
        QuantizableRhythmFeatureExtractorMixin.__init__(self, quantize_enabled)
        self._cyclic = None
        self.cyclic = cyclic  # call setters

    @staticmethod
    def init_pre_processors():
        yield IOIVector(mode=IOIVector.Mode.POST_NOTE)

    @property
    def cyclic(self):
        """Cyclic behaviour, see process documentation"""
        return self._cyclic

    @cyclic.setter
    def cyclic(self, cyclic):
        self._cyclic = bool(cyclic)

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Returns the interval difference vector (aka difference of rhythm vector) of the given rhythm. Per note, this is
        the difference between the current onset interval and the next onset interval. So, if N is the number of onsets,
        the returned vector will have a length of N - 1. This is different with cyclic rhythms, where the last onset's
        interval is compared with the first onset's interval. In this case, the length will be N.
        The inter-onset interval vector is computed in POST_NOTE mode.

        For example, given the POST_NOTE inter-onset interval vector for the Rumba clave:
          [3, 4, 3, 2, 4]

        The interval difference vector would be:
           With cyclic set to False: [4/3, 3/4, 2/3, 4/2]
           With cyclic set to True:  [4/3, 3/4, 2/3, 4/2, 3/4]

        :param rhythm: monophonic rhythm of which to compute the inter-onset difference vector
        :return: interval difference vector of the given rhythm
        """

        vector, i = pre_processor_results[0], 0

        if self.cyclic:
            vector.append(vector[0])

        while i < len(vector) - 1:
            try:
                vector[i] = vector[i + 1] / float(vector[i])
            except ZeroDivisionError:
                vector[i] = float('inf')
            i += 1

        vector.pop()
        return vector


class OnsetPositionVector(MonophonicRhythmFeatureExtractor, QuantizableRhythmFeatureExtractorMixin):
    def __init__(self, unit: tp.Optional[UnitType] = Unit.EIGHTH, quantize_enabled=True):
        MonophonicRhythmFeatureExtractor.__init__(self, unit)
        QuantizableRhythmFeatureExtractorMixin.__init__(self, quantize_enabled)

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Returns the absolute onset times of the notes in the given rhythm.

        :return: a list with the onset times of this rhythm
        """

        Rhythm.Precondition.check_resolution(rhythm)
        resolution = rhythm.get_resolution()
        quantize = self.is_quantize_enabled()
        unit = self.get_unit()

        return [convert_tick(onset[0], resolution, unit, quantize) for onset in rhythm.get_onsets()]


class SyncopationVector(MonophonicRhythmFeatureExtractor):
    def __init__(self, unit: UnitType = Unit.EIGHTH, equal_upbeat_salience_profile: bool = False):
        super().__init__(unit)
        self.equal_upbeat_salience_profile = bool(equal_upbeat_salience_profile)

    @staticmethod
    def init_pre_processors():
        yield NoteVector(ret_positions=True)

    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        if unit is None:
            raise ValueError("SyncopationVector does not support tick-based computation")
        super().set_unit(unit)

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Extracts the syncopations from the given monophonic rhythm. The syncopations are computed with the method
        proposed by H.C. Longuet-Higgins and C. S. Lee in their work titled: "The Rhythmic Interpretation of
        Monophonic Music".

        The syncopations are returned as a sequence of tuples containing three elements:
            - the syncopation strength
            - the syncopated note position
            - the position of the rest against which the note is syncopated

        :param rhythm: rhythm from which to extract the syncopations
        :return: the syncopations as (syncopation strength, note position, rest position) tuples
        """

        Rhythm.Precondition.check_time_signature(rhythm)

        note_vector = pre_processor_results[0]
        time_signature = rhythm.get_time_signature()
        metrical_weights = time_signature.get_salience_profile(
            self.unit, equal_upbeats=self.equal_upbeat_salience_profile)

        for n, [curr_event_type, _, curr_step] in enumerate(note_vector):
            next_event_type, next_event_duration, next_step = note_vector[(n + 1) % len(note_vector)]

            # only iterate over note-rest pairs
            if not (curr_event_type == NoteVector.NOTE and next_event_type == NoteVector.REST):
                continue

            note_weight = metrical_weights[curr_step % len(metrical_weights)]
            rest_weight = metrical_weights[next_step % len(metrical_weights)]

            if rest_weight >= note_weight:
                syncopation_strength = rest_weight - note_weight
                yield syncopation_strength, curr_step, next_step


class SyncopatedOnsetRatio(MonophonicRhythmFeatureExtractor):
    def __init__(self, unit: UnitType = Unit.EIGHTH, ret_fraction: bool = False):
        super().__init__(unit)
        self._ret_fraction = None
        self.ret_fraction = ret_fraction  # calling setter

    @staticmethod
    def init_pre_processors():
        # NOTE: Order is important (__process__ depends on it)
        yield BinaryOnsetVector()
        yield SyncopationVector()

    @property
    def ret_fraction(self) -> bool:
        return self._ret_fraction

    @ret_fraction.setter
    def ret_fraction(self, ret_fraction):
        self._ret_fraction = bool(ret_fraction)

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Returns the number of syncopated onsets over the total number of onsets. The syncopations are computed
        SyncopationVector.

        :param rhythm: rhythm for which to compute the syncopated onset ratio
        :return: either if ret_fraction is True a Fraction object, otherwise a float
        """

        binary_onset_vector = pre_processor_results[0]
        syncopation_vector = pre_processor_results[1]

        # NOTE: We use binary onset vector and don't call rhythm.get_onset_count because the onset count might be
        # different for different units (onsets might merge with large unit like Unit.QUARTER)
        n_onsets = sum(binary_onset_vector)
        n_syncopated_onsets = len(syncopation_vector)
        assert n_syncopated_onsets <= n_onsets

        if self.ret_fraction:
            return Fraction(n_syncopated_onsets, n_onsets)
        else:
            return n_syncopated_onsets / float(n_onsets)


class MeanSyncopationStrength(MonophonicRhythmFeatureExtractor):
    @staticmethod
    def init_pre_processors():
        yield SyncopationVector()

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Returns the average syncopation strength per step. The step size depends on the unit (see set_unit). The
        syncopations are computed with SyncopationVector.

        :param rhythm: rhythm of which to compute the average syncopation strength
        :return: average syncopation strength per step
        """

        syncopation_vector = pre_processor_results[0]
        total_syncopation_strength = sum(info[1] for info in syncopation_vector)
        n_steps = rhythm.get_duration(self.unit, ceil=True)

        try:
            return total_syncopation_strength / n_steps
        except ZeroDivisionError:
            return 0


class OnsetDensity(MonophonicRhythmFeatureExtractor):
    def __init__(self, unit: tp.Optional[UnitType] = Unit.EIGHTH):
        super().__init__(unit)

    @staticmethod
    def init_pre_processors():
        yield BinaryOnsetVector()

    def __process__(self, rhythm: MonophonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Computes the onset density of the given rhythm. The onset density is the number of onsets over the number of
        positions in the binary onset vector of the given rhythm.

        :param rhythm: monophonic rhythm to compute the onset density of
        :return: onset density of the given rhythm
        """

        binary_vector = pre_processor_results[0]
        n_onsets = sum(binary_vector)  # onsets are ones, non-onsets are zeros
        return float(n_onsets) / len(binary_vector)


#######################################################
# Polyphonic rhythm feature extractor implementations #
#######################################################


class MultiChannelMonophonicRhythmFeatureVector(PolyphonicRhythmFeatureExtractor):
    class MonophonicFeatureExtractorNotSet(Exception):
        pass

    def __init__(
            self, unit: tp.Optional[UnitType] = None,
            mono_extractor_cls: tp.Optional[tp.Type[MonophonicRhythmFeatureExtractor]] = None,
            *mono_extractor_args,
            **mono_extractor_kwargs
    ):
        # TODO if possible, "inherit" the default unit of the mono extractor
        self._mono_extractor = None  # type: tp.Union[MonophonicRhythmFeatureExtractor, None]
        super().__init__(unit)

        if mono_extractor_cls:
            mono_extractor = mono_extractor_cls(unit, *mono_extractor_args, **mono_extractor_kwargs)
            self.monophonic_extractor = mono_extractor

    @property
    def monophonic_extractor(self):
        return self._mono_extractor

    @monophonic_extractor.setter
    def monophonic_extractor(self, monophonic_feature_extractor: MonophonicRhythmFeatureExtractor):
        monophonic_feature_extractor.set_unit(self.unit)  # TODO Shouldn't this one adapt to the mono extractor's unit?
        self._mono_extractor = monophonic_feature_extractor

    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        super().set_unit(unit)
        if self._mono_extractor:
            self._mono_extractor.set_unit(unit)

    def __process__(self, rhythm: PolyphonicRhythm, pre_processor_results: tp.List[tp.Any]):
        if not self._mono_extractor:
            raise self.MonophonicFeatureExtractorNotSet("Please set the mono_extractor property")
        return OrderedDict(self.__compute_features_per_track(rhythm))

    def __compute_features_per_track(self, rhythm: PolyphonicRhythm) -> tp.Generator[tp.Tuple[str, tp.Any], None, None]:
        for track in rhythm.get_track_iterator():
            yield track.get_name(), self._mono_extractor.process(track)


class PolyphonicSyncopationVector(PolyphonicRhythmFeatureExtractor):
    def __init__(self, unit: UnitType = Unit.EIGHTH, equal_upbeat_salience_profile: bool = False):
        super().__init__(unit)
        self.equal_upbeat_salience_profile = bool(equal_upbeat_salience_profile)
        self._f_get_instrumentation_weight = None
        # type: tp.Optional[tp.Callable[[tp.Set[str], tp.Set[str]], tp.Union[int, float]]]

    @staticmethod
    def init_pre_processors():
        multi_channel_note_vector = MultiChannelMonophonicRhythmFeatureVector(Unit.EIGHTH)
        multi_channel_note_vector.monophonic_extractor = NoteVector(ret_positions=True)
        yield multi_channel_note_vector

    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        if unit is None:
            raise ValueError("PolyphonicSyncopationVector does not support tick-based computation")
        super().set_unit(unit)

    def set_instrumentation_weight_function(
            self,
            func: tp.Union[tp.Callable[[tp.Set[str], tp.Set[str]], tp.Union[int, float]], None]
    ):
        if func and not callable(func):
            raise TypeError

        self._f_get_instrumentation_weight = func or None

    def get_instrumentation_weight_function(self) -> \
            tp.Union[tp.Callable[[tp.Set[str], tp.Set[str]], tp.Union[int, float]], None]:
        return self._f_get_instrumentation_weight

    # noinspection PyUnusedLocal
    @staticmethod
    def default_instrumentation_weight_function(
            syncopated_instruments: tp.Set[str],
            other_instruments: tp.Set[str]
    ) -> int:
        n_other_instruments = len(other_instruments)
        return min(3 - n_other_instruments, 0)

    def __process__(self, rhythm: PolyphonicRhythm, pre_processor_results: tp.List[tp.Any]):
        """
        Finds the polyphonic syncopations and their syncopation strengths. This is an implementation of the method
        proposed by Maria A.G. Witek et al in their work titled "Syncopation, Body-Movement and Pleasure in Groove
        Music".

        The definition of syncopation, as proposed in the work of Witek, goes as follows:

            If N is a note that precedes a rest, R, and R has a metric weight greater than or equal to N, then the
            pair (N, R) is said to constitute a monophonic syncopation. If N is a note on a certain instrument that
            precedes a note on a different instrument, Ndi, and Ndi has a metric weight greater than or equal to N,
            then the pair (N, Ndi) is said to constitute a polyphonic syncopation.

        This definition is used to find the syncopations. Then, the syncopation strengths are computed with this
        formula:

            S = Ndi - N + I

        where S is the degree of syncopation, Ndi is the metrical weight of the note succeeding the syncopated note, N
        is the metrical weight of the syncopated note and I is the instrumentation weight. The instrumentation weight
        depends on the relation of the instruments involved in the syncopation. This relation depends on two factors:

            - the number of instruments involved in the syncopation
            - the type of instruments involved in the syncopation

        As there is no formula given by Witek for how to compute this value, the computation of this value is up to the
        owner of this feature extractor. The function to compute the weight can be set through the
        set_instrumentation_weight_function method and should be a callable receiving two arguments:

            - the names of the tracks (instruments) that play a syncopated note
            - the names of the tracks (instruments) that the note is syncopated against (empty if syncopated against a
              rest)

        The syncopations are returned as a sequence of three-element tuples containing:
            - the degree of syncopation (syncopation strength)
            - position of the syncopated note(s)
            - position of the note(s)/rest(s) against which the note(s) are syncopated

        The syncopations are returned (position, syncopation strength) tuples.

        NOTE: the formula in the Witek's work is different: S = N - Ndi + I. I suspect that it is a typo, as examples
        in the same work show that the formula, S = Ndi - N + I, is used.

        :param rhythm: the polyphonic rhythm of which to compute the polyphonic syncopation vector
        :return: syncopations as (syncopation strength, syncopated event position, other event position) tuples
        """

        Rhythm.Precondition.check_time_signature(rhythm)
        note_vectors = pre_processor_results[0]

        if not note_vectors:
            return

        time_signature = rhythm.get_time_signature()
        salience_profile = time_signature.get_salience_profile(self.unit, self.equal_upbeat_salience_profile)
        get_instrumentation_weight = self._f_get_instrumentation_weight or self.default_instrumentation_weight_function

        # A dictionary containing (instrument, event) tuples by event position
        instrument_event_pairs_by_position = defaultdict(lambda: set())
        for instrument, events in note_vectors.items():
            # NoteVector with ret_positions set to True adds event position as 3rd element (= e[2])
            for e in events:
                instrument_event_pairs_by_position[e[2]].add((instrument, e))

        # Positions that contain an event (either a rest or a note) in ascending order
        dirty_positions = sorted(instrument_event_pairs_by_position.keys())
        assert dirty_positions[0] == 0, "first step should always contain an event (either a rest or a note)"

        # Iterate over the positions that contain events
        for i_position, curr_step in enumerate(dirty_positions):
            next_step = dirty_positions[(i_position + 1) % len(dirty_positions)]
            curr_metrical_weight = salience_profile[curr_step % len(salience_profile)]
            next_metrical_weight = salience_profile[next_step % len(salience_profile)]

            # Syncopation not possible on this current metrical position pair
            if curr_metrical_weight > next_metrical_weight:
                continue

            curr_instrument_event_pairs = instrument_event_pairs_by_position[curr_step]
            next_instrument_event_pairs = instrument_event_pairs_by_position[next_step]

            curr_sounding_inst_set = set(inst for inst, evt in curr_instrument_event_pairs if evt[0] == NoteVector.NOTE)
            next_sounding_inst_set = set(inst for inst, evt in next_instrument_event_pairs if evt[0] == NoteVector.NOTE)

            if len(curr_sounding_inst_set) < 0:
                continue

            # Detect a syncopation if there is at least one instrument that plays a sounding note current position and a
            # rest (or a tied note) at next position
            if any(inst not in next_sounding_inst_set for inst in curr_sounding_inst_set):
                instrumentation_weight = get_instrumentation_weight(curr_sounding_inst_set, next_sounding_inst_set)
                syncopation_degree = next_metrical_weight - curr_metrical_weight + instrumentation_weight
                yield syncopation_degree, curr_step, next_step


__all__ = [
    # Feature extractor abstract base classes (or interfaces)
    'FeatureExtractor', 'RhythmFeatureExtractor',
    'MonophonicRhythmFeatureExtractor', 'PolyphonicRhythmFeatureExtractor',

    # Monophonic rhythm feature extractor implementations
    'BinaryOnsetVector', 'NoteVector', 'IOIVector', 'IOIHistogram', 'BinarySchillingerChain', 'ChronotonicChain',
    'IOIDifferenceVector', 'OnsetPositionVector', 'SyncopationVector', 'SyncopatedOnsetRatio',
    'MeanSyncopationStrength', 'OnsetDensity',

    # Polyphonic rhythm feature extractor implementations
    'MultiChannelMonophonicRhythmFeatureVector', 'PolyphonicSyncopationVector'
]



import typing as tp
from fractions import Fraction
from unittest import TestCase, main
from unittest.mock import MagicMock, PropertyMock, patch
from collections import OrderedDict
from abc import ABCMeta, abstractmethod

# feature extractor base classes
from beatsearch.feature_extraction import FeatureExtractor, RhythmFeatureExtractor, \
    MonophonicRhythmFeatureExtractor, PolyphonicRhythmFeatureExtractor

# monophonic feature extractor implementations
from beatsearch.feature_extraction import (
    BinaryOnsetVector,
    NoteVector,
    BinarySchillingerChain,
    ChronotonicChain,
    OnsetDensity,
    OnsetPositionVector,
    SyncopationVector,
    SyncopatedOnsetRatio,
    MeanSyncopationStrength,
    IOIVector,
    IOIDifferenceVector,
    IOIHistogram
)

from beatsearch.feature_extraction import (
    MultiChannelMonophonicRhythmFeatureVector
)

# misc
from beatsearch.rhythm import Rhythm, Track, MonophonicRhythm, PolyphonicRhythm, Unit, TimeSignature
from beatsearch.test_rhythm import mock_onset


def set_rhythm_mock_properties(rhythm_mock, resolution, time_signature, duration_in_ticks):
    # resolution
    rhythm_mock.get_resolution.return_value = resolution
    rhythm_mock.resolution = resolution
    # time signature
    rhythm_mock.get_time_signature.return_value = time_signature
    rhythm_mock.time_signature = time_signature
    # duration (in ticks)
    rhythm_mock.get_duration_in_ticks.return_value = duration_in_ticks
    rhythm_mock.duration_in_ticks = duration_in_ticks
    rhythm_mock.get_duration.return_value = duration_in_ticks


def get_mono_rhythm_mock_with_23_rumba_clave_onsets(resolution):
    rhythm_mock = MagicMock(MonophonicRhythm)
    assert resolution >= 4, "the 2/3 claves pattern is is not representable with a resolution smaller than 4"

    onset_positions = (
        int(resolution / 4.0 * 0),
        int(resolution / 4.0 * 3),
        int(resolution / 4.0 * 7),
        int(resolution / 4.0 * 10),
        int(resolution / 4.0 * 12)
    )

    mocked_onsets = tuple(mock_onset(tick, 100) for tick in onset_positions)
    rhythm_mock.get_onsets.return_value = mocked_onsets
    rhythm_mock.onsets = rhythm_mock.get_onsets.return_value

    set_rhythm_mock_properties(rhythm_mock, resolution, (4, 4), int(resolution) * 4)
    return rhythm_mock


def get_mocked_track(name, onset_positions, resolution, time_signature, duration_in_ticks):
    track_mock = MagicMock(Track)
    mocked_onsets = tuple(mock_onset(tick, 100) for tick in onset_positions)

    track_mock.get_name.return_value = name
    track_mock.name = name
    track_mock.get_onsets.return_value = mocked_onsets
    track_mock.onsets = mocked_onsets

    set_rhythm_mock_properties(track_mock, resolution, time_signature, duration_in_ticks)
    return track_mock


def get_poly_rhythm_mock_with_songo_onsets(resolution):
    rhythm_mock = MagicMock(PolyphonicRhythm)
    assert resolution >= 4, "the songo rhythm is not representable with a resolution smaller than 4"
    ts = (4, 4)
    duration_in_ticks = 4 * resolution  # one measure 4/4

    kick_track_mock = get_mocked_track("kick", [
        int(resolution / 4.0 * 3),
        int(resolution / 4.0 * 6),
        int(resolution / 4.0 * 11),
        int(resolution / 4.0 * 14)
    ], resolution, ts, duration_in_ticks)

    snare_track_mock = get_mocked_track("snare", [
        int(resolution / 4.0 * 2),
        int(resolution / 4.0 * 5),
        int(resolution / 4.0 * 7),
        int(resolution / 4.0 * 9),
        int(resolution / 4.0 * 10),
        int(resolution / 4.0 * 13),
        int(resolution / 4.0 * 15)
    ], resolution, ts, duration_in_ticks)

    hihat_track_mock = get_mocked_track("hihat", [
        int(resolution / 4.0 * 0),
        int(resolution / 4.0 * 4),
        int(resolution / 4.0 * 8),
        int(resolution / 4.0 * 12)
    ], resolution, ts, duration_in_ticks)

    rhythm_mock.get_track_iterator.return_value = iter([kick_track_mock, snare_track_mock, hihat_track_mock])
    set_rhythm_mock_properties(rhythm_mock, resolution, ts, duration_in_ticks)
    return rhythm_mock


class TestFeatureExtractor(TestCase):
    def test_not_instantiable(self):
        self.assertRaises(Exception, FeatureExtractor)


class TestRhythmFeatureExtractor(TestCase):
    def test_not_instantiable(self):
        self.assertRaises(Exception, RhythmFeatureExtractor)


class TestMonophonicRhythmFeatureExtractor(TestCase):
    def test_not_instantiable(self):
        self.assertRaises(Exception, MonophonicRhythmFeatureExtractor)


class TestRhythmFeatureExtractorImplementationMixin(object, metaclass=ABCMeta):
    def test_instantiable(self):
        cls = self.get_impl_class()
        cls()  # shouldn't cause any problems

    # noinspection PyUnresolvedReferences
    def test_unit_set_with_first_positional_constructor_argument(self):
        cls = self.get_impl_class()
        for unit in self.get_legal_units():
            with self.subTest(unit):
                obj = cls(unit)
                self.assertEqual(obj.unit, unit)

    # noinspection PyUnresolvedReferences
    def test_unit_set_with_named_constructor_argument(self):
        cls = self.get_impl_class()
        for unit in self.get_legal_units():
            with self.subTest(unit):
                obj = cls(unit=unit)
                self.assertEqual(obj.unit, unit)

    # noinspection PyUnresolvedReferences
    def test_unit_set_to_pre_processors_with_first_positional_constructor_argument(self):
        cls = self.get_impl_class()
        for unit in self.get_legal_units():
            obj = cls(unit=unit)
            for pre_processor in obj.pre_processors:
                with self.subTest("%s.%s" % (unit, pre_processor.__class__.__name__)):
                    self.assertEqual(pre_processor.unit, unit)

    # noinspection PyUnresolvedReferences
    def test_unit_set_to_preprocessors_with_named_constructor_argument(self):
        cls = self.get_impl_class()
        for unit in self.get_legal_units():
            obj = cls(unit=unit)
            for pre_processor in obj.pre_processors:
                with self.subTest("%s.%s" % (unit, pre_processor.__class__.__name__)):
                    self.assertEqual(pre_processor.unit, unit)

    # noinspection PyUnresolvedReferences
    def test_unit_property_sets_preprocessor_units(self):
        cls = self.get_impl_class()
        obj = cls()
        for unit in self.get_legal_units():
            obj.unit = unit
            for pre_processor in obj.pre_processors:
                with self.subTest("%s.%s" % (unit, pre_processor.__class__.__name__)):
                    self.assertEqual(pre_processor.unit, unit)

    @staticmethod
    @abstractmethod
    def get_impl_class() -> tp.Type[RhythmFeatureExtractor]:
        raise NotImplementedError


#######################################################
# Monophonic rhythm feature extractor implementations #
#######################################################


class TestMonophonicRhythmFeatureExtractorImplementationMixin(TestRhythmFeatureExtractorImplementationMixin,
                                                              metaclass=ABCMeta):
    def __init__(self, *args, **kw):
        # noinspection PyArgumentList
        super().__init__(*args, **kw)
        self.rhythm = get_mono_rhythm_mock_with_23_rumba_clave_onsets(4)  # type: MonophonicRhythm
        self.feature_extractor = None  # type: tp.Union[MonophonicRhythmFeatureExtractor, None]

    # noinspection PyPep8Naming
    def setUp(self):
        cls = self.get_impl_class()
        self.feature_extractor = cls()
        self.feature_extractor.unit = 1/16

    @staticmethod
    def get_legal_units():
        # noinspection PyTypeChecker
        return [None] + list(Unit)

    @staticmethod
    @abstractmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        raise NotImplementedError


class TestBinaryOnsetVector(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return BinaryOnsetVector

    def test_process(self):
        expected_binary_ticks = [1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0]
        actual_binary_ticks = self.feature_extractor.process(self.rhythm)
        self.assertEqual(actual_binary_ticks, expected_binary_ticks)


class TestNoteVector(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):

    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return NoteVector

    @staticmethod
    def get_legal_units():
        # noinspection PyTypeChecker
        # Syncopation vector does not support tick-based computation
        return list(Unit)

    def test_process(self):
        expected_note_vector_string = "N2 R1 N1 R2 R1 N1 R2 N2 N4"
        to_event = lambda c: [NoteVector.NOTE, NoteVector.REST]["NR".index(c)]

        # TODO mock time signature and its methods
        rhythm = self.rhythm
        rhythm.time_signature = TimeSignature(4, 4)
        rhythm.get_time_signature.return_value = rhythm.time_signature

        expected_note_vector = tuple((to_event(s[0]), int(s[1])) for s in expected_note_vector_string.split())
        actual_note_vector = self.feature_extractor.process(rhythm)
        self.assertSequenceEqual(actual_note_vector, expected_note_vector)


class TestIOIVector(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class():
        return IOIVector

    def test_process_in_pre_note_mode(self):
        extractor = self.feature_extractor  # type: IOIVector
        extractor.mode = IOIVector.Mode.PRE_NOTE

        expected_ioi_vector = [0, 3, 4, 3, 2]
        actual_ioi_vector = extractor.process(self.rhythm)
        self.assertEqual(actual_ioi_vector, expected_ioi_vector)

    def test_process_in_post_note_mode(self):
        extractor = self.feature_extractor  # type: IOIVector
        extractor.mode = IOIVector.Mode.POST_NOTE

        expected_ioi_vector = [3, 4, 3, 2, 4]
        actual_ioi_vector = extractor.process(self.rhythm)
        self.assertEqual(actual_ioi_vector, expected_ioi_vector)


class TestIOIHistogram(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return IOIHistogram

    def test_process(self):
        expected_histogram = (
            [1, 2, 2],  # occurrences
            [2, 3, 4]  # intervals
        )
        actual_histogram = self.feature_extractor.process(self.rhythm)
        self.assertEqual(actual_histogram, expected_histogram)


class TestBinarySchillingerChain(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return BinarySchillingerChain

    def test_process_with_values_one_zero(self):
        extractor = self.feature_extractor  # type: BinarySchillingerChain
        extractor.values = (1, 0)

        expected_chain = [1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 1]
        actual_chain = extractor.process(self.rhythm)
        self.assertEqual(actual_chain, expected_chain)

    def test_process_with_values_zero_one(self):
        extractor = self.feature_extractor  # type: BinarySchillingerChain
        extractor.values = (0, 1)

        expected_chain = [0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0]
        actual_chain = extractor.process(self.rhythm)
        self.assertEqual(actual_chain, expected_chain)


class TestChronotonicChain(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return ChronotonicChain

    def test_process(self):
        expected_chain = [3, 3, 3, 4, 4, 4, 4, 3, 3, 3, 2, 2, 4, 4, 4, 4]
        actual_chain = self.feature_extractor.process(self.rhythm)
        self.assertEqual(actual_chain, expected_chain)


class TestIOIDifferenceVector(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return IOIDifferenceVector

    def test_process_in_non_cyclic_mode(self):
        extractor = self.feature_extractor  # type: IOIDifferenceVector
        extractor.cyclic = False

        expected_vector = [4 / 3, 3 / 4, 2 / 3, 4 / 2]
        actual_vector = extractor.process(self.rhythm)
        self.assertEqual(actual_vector, expected_vector)

    def test_process_in_cyclic_mode(self):
        extractor = self.feature_extractor  # type: IOIDifferenceVector
        extractor.cyclic = True

        expected_vector = [4 / 3, 3 / 4, 2 / 3, 4 / 2, 3 / 4]
        actual_vector = extractor.process(self.rhythm)
        self.assertEqual(actual_vector, expected_vector)


class TestOnsetPositionVector(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return OnsetPositionVector

    def test_process(self):
        expected_vector = [0, 3, 7, 10, 12]
        actual_vector = self.feature_extractor.process(self.rhythm)
        self.assertEqual(actual_vector, expected_vector)


class TestSyncopationVector(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return SyncopationVector

    @staticmethod
    def get_legal_units():
        # noinspection PyTypeChecker
        # Syncopation vector does not support tick-based computation
        return list(Unit)

    # TODO Add test_process


class TestSyncopatedOnsetRatio(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return SyncopatedOnsetRatio

    @staticmethod
    def get_legal_units():
        return TestSyncopationVector.get_legal_units()

    def test_default_ret_float(self):
        self.assertFalse(self.feature_extractor.ret_fraction, "should return a float by default")

    @patch.object(SyncopationVector, "process")
    def test_process_ret_fraction(self, mock_syncopation_vector_process):
        syncopations = "first", "second", "third"
        mock_syncopation_vector_process.return_value = syncopations

        extractor = self.feature_extractor  # type: SyncopatedOnsetRatio
        extractor.ret_fraction = True

        expected_ratio = Fraction(len(syncopations), 5)  # self.rhythm contains 5 onsets
        actual_ratio = extractor.process(self.rhythm)
        self.assertEqual(actual_ratio, expected_ratio)

    @patch.object(SyncopationVector, "process")
    def test_process_ret_float(self, mock_syncopation_vector_process):
        syncopations = "first", "second", "third"
        mock_syncopation_vector_process.return_value = syncopations

        extractor = self.feature_extractor  # type: SyncopatedOnsetRatio
        extractor.ret_fraction = False

        expected_ratio = 3 / float(5)
        actual_ratio = extractor.process(self.rhythm)
        self.assertAlmostEqual(actual_ratio, expected_ratio)


class TestMeanSyncopationStrength(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return MeanSyncopationStrength

    @staticmethod
    def get_legal_units():
        return TestSyncopationVector.get_legal_units()

    @patch.object(SyncopationVector, "process")
    def test_process(self, mock_syncopation_vector_process):
        syncopations = [0, 7], [1, 2], [2, 5]  # total sync strength = 14
        mock_syncopation_vector_process.return_value = syncopations
        self.rhythm.get_duration.return_value = 123

        expected_mean_sync_strength = 14 / 123
        actual_mean_sync_strength = self.feature_extractor.process(self.rhythm)
        self.assertEqual(actual_mean_sync_strength, expected_mean_sync_strength)


class TestOnsetDensity(TestMonophonicRhythmFeatureExtractorImplementationMixin, TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[MonophonicRhythmFeatureExtractor]:
        return OnsetDensity

    def test_process(self):
        expected_onset_density = 0.3125  # rumba rhythm has 5 onsets and 16 available positions
        actual_onset_density = self.feature_extractor.process(self.rhythm)
        self.assertEqual(actual_onset_density, expected_onset_density)


class TestPolyphonicRhythmFeatureExtractor(TestCase):
    def test_not_instantiable(self):
        self.assertRaises(Exception, PolyphonicRhythmFeatureExtractor)


#######################################################
# Polyphonic rhythm feature extractor implementations #
#######################################################

class TestPolyphonicRhythmFeatureExtractorImplementationMixin(TestRhythmFeatureExtractorImplementationMixin,
                                                              metaclass=ABCMeta):
    def __init__(self, *args, **kw):
        # noinspection PyArgumentList
        super().__init__(*args, **kw)
        self.rhythm = get_poly_rhythm_mock_with_songo_onsets(4)  # type: PolyphonicRhythm
        self.feature_extractor = None  # type: tp.Union[PolyphonicRhythmFeatureExtractor, None]

    # noinspection PyPep8Naming
    def setUp(self):
        cls = self.get_impl_class()
        self.feature_extractor = cls()
        self.feature_extractor.unit = 1/16

    @staticmethod
    def get_legal_units():
        # noinspection PyTypeChecker
        return [None] + list(Unit)

    @staticmethod
    @abstractmethod
    def get_impl_class() -> tp.Type[PolyphonicRhythmFeatureExtractor]:
        raise NotImplementedError


class TestMultiChannelMonophonicRhythmFeatureVector(TestPolyphonicRhythmFeatureExtractorImplementationMixin,
                                                    TestCase):
    @staticmethod
    def get_impl_class() -> tp.Type[PolyphonicRhythmFeatureExtractor]:
        return MultiChannelMonophonicRhythmFeatureVector

    def test_process(self):
        poly_extractor = self.feature_extractor  # type: MultiChannelMonophonicRhythmFeatureVector

        mono_extractor_mock = MagicMock(MonophonicRhythmFeatureExtractor)
        mono_extractor_mock.process.side_effect = ["fake_feature_a", "fake_feature_b", "fake_feature_c"]

        expected_result_items = (
            ("kick", "fake_feature_a"),
            ("snare", "fake_feature_b"),
            ("hihat", "fake_feature_c")
        )

        poly_extractor.monophonic_extractor = mono_extractor_mock
        actual_result_items = tuple(poly_extractor.process(self.rhythm).items())
        self.assertEqual(actual_result_items, expected_result_items)


# TODO: Add tests for polyphonic syncopation vector


if __name__ == "__main__":
    main()



import math
from six import with_metaclass


class EventRegistry(object):
    Events = {}
    MetaEvents = {}

    def register_event(cls, event, bases):
        if (Event in bases) or (NoteEvent in bases):
            assert event.statusmsg not in cls.Events, \
                "Event %s already registered" % event.name
            cls.Events[event.statusmsg] = event
        elif (MetaEvent in bases) or (MetaEventWithText in bases):
            if event.metacommand is not None:
                assert event.metacommand not in cls.MetaEvents, \
                    "Event %s already registered" % event.name
                cls.MetaEvents[event.metacommand] = event
        else:
            raise ValueError(
                "Unknown bases class in event type: " + event.name)
    register_event = classmethod(register_event)


class AbstractEventMetaclass(type):

    def __init__(cls, name, bases, dict):
        if name not in ['AbstractEvent', 'Event', 'MetaEvent', 'NoteEvent',
                        'MetaEventWithText']:
            EventRegistry.register_event(cls, bases)


class AbstractEvent(with_metaclass(AbstractEventMetaclass, object)):
    # __slots__ = ['tick', 'data']
    name = "Generic MIDI Event"
    length = 0
    statusmsg = 0x0

    def __init__(self, **kw):
        if type(self.length) == int:
            defdata = [0] * self.length
        else:
            defdata = []
        self.tick = 0
        self.data = defdata
        for key in kw:
            setattr(self, key, kw[key])

    def __cmp__(self, other):
        if self.tick < other.tick:
            return -1
        elif self.tick > other.tick:
            return 1
        return cmp(self.data, other.data)

    def __baserepr__(self, keys=[]):
        keys = ['tick'] + keys + ['data']
        body = []
        for key in keys:
            val = getattr(self, key)
            keyval = "%s=%r" % (key, val)
            body.append(keyval)
        body = str.join(', ', body)
        return "midi.%s(%s)" % (self.__class__.__name__, body)

    def __repr__(self):
        return self.__baserepr__()


class Event(AbstractEvent):
    # __slots__ = ['channel']
    name = 'Event'

    def __init__(self, **kw):
        if 'channel' not in kw:
            kw = kw.copy()
            kw['channel'] = 0
        super(Event, self).__init__(**kw)

    def copy(self, **kw):
        _kw = {'channel': self.channel, 'tick': self.tick, 'data': self.data}
        _kw.update(kw)
        return self.__class__(**_kw)

    def __cmp__(self, other):
        if self.tick < other.tick:
            return -1
        elif self.tick > other.tick:
            return 1
        return 0

    def __repr__(self):
        return self.__baserepr__(['channel'])

    def is_event(cls, statusmsg):
        return (cls.statusmsg == (statusmsg & 0xF0))
    is_event = classmethod(is_event)


"""
MetaEvent is a special subclass of Event that is not meant to
be used as a concrete class.  It defines a subset of Events known
as the Meta events.
"""


class MetaEvent(AbstractEvent):
    statusmsg = 0xFF
    metacommand = 0x0
    name = 'Meta Event'

    def is_event(cls, statusmsg):
        return (statusmsg == 0xFF)
    is_event = classmethod(is_event)


"""
NoteEvent is a special subclass of Event that is not meant to
be used as a concrete class.  It defines the generalities of NoteOn
and NoteOff events.
"""


class NoteEvent(Event):
    # __slots__ = ['pitch', 'velocity']
    length = 2

    def get_pitch(self):
        return self.data[0]

    def set_pitch(self, val):
        self.data[0] = val
    pitch = property(get_pitch, set_pitch)

    def get_velocity(self):
        return self.data[1]

    def set_velocity(self, val):
        self.data[1] = val
    velocity = property(get_velocity, set_velocity)


class NoteOnEvent(NoteEvent):
    statusmsg = 0x90
    name = 'Note On'


class NoteOffEvent(NoteEvent):
    statusmsg = 0x80
    name = 'Note Off'


class AfterTouchEvent(Event):
    statusmsg = 0xA0
    length = 2
    name = 'After Touch'

    def get_pitch(self):
        return self.data[0]

    def set_pitch(self, val):
        self.data[0] = val
    pitch = property(get_pitch, set_pitch)

    def get_value(self):
        return self.data[1]

    def set_value(self, val):
        self.data[1] = val
    value = property(get_value, set_value)


class ControlChangeEvent(Event):
    # __slots__ = ['control', 'value']
    statusmsg = 0xB0
    length = 2
    name = 'Control Change'

    def set_control(self, val):
        self.data[0] = val

    def get_control(self):
        return self.data[0]
    control = property(get_control, set_control)

    def set_value(self, val):
        self.data[1] = val

    def get_value(self):
        return self.data[1]
    value = property(get_value, set_value)


class ProgramChangeEvent(Event):
    # __slots__ = ['value']
    statusmsg = 0xC0
    length = 1
    name = 'Program Change'

    def set_value(self, val):
        self.data[0] = val

    def get_value(self):
        return self.data[0]
    value = property(get_value, set_value)


class ChannelAfterTouchEvent(Event):
    # __slots__ = ['value']
    statusmsg = 0xD0
    length = 1
    name = 'Channel After Touch'

    def set_value(self, val):
        self.data[1] = val

    def get_value(self):
        return self.data[1]
    value = property(get_value, set_value)


class PitchWheelEvent(Event):
    # __slots__ = ['pitch']
    statusmsg = 0xE0
    length = 2
    name = 'Pitch Wheel'

    def get_pitch(self):
        return ((self.data[1] << 7) | self.data[0]) - 0x2000

    def set_pitch(self, pitch):
        value = pitch + 0x2000
        self.data[0] = value & 0x7F
        self.data[1] = (value >> 7) & 0x7F
    pitch = property(get_pitch, set_pitch)


class SysexEvent(Event):
    statusmsg = 0xF0
    name = 'SysEx'
    length = 'varlen'

    def is_event(cls, statusmsg):
        return (cls.statusmsg == statusmsg)
    is_event = classmethod(is_event)


class SequenceNumberMetaEvent(MetaEvent):
    name = 'Sequence Number'
    metacommand = 0x00
    length = 2


class MetaEventWithText(MetaEvent):

    def __init__(self, **kw):
        super(MetaEventWithText, self).__init__(**kw)
        if 'text' not in kw:
            self.text = ''.join(chr(datum) for datum in self.data)

    def __repr__(self):
        return self.__baserepr__(['text'])


class TextMetaEvent(MetaEventWithText):
    name = 'Text'
    metacommand = 0x01
    length = 'varlen'


class CopyrightMetaEvent(MetaEventWithText):
    name = 'Copyright Notice'
    metacommand = 0x02
    length = 'varlen'


class TrackNameEvent(MetaEventWithText):
    name = 'Track Name'
    metacommand = 0x03
    length = 'varlen'


class InstrumentNameEvent(MetaEventWithText):
    name = 'Instrument Name'
    metacommand = 0x04
    length = 'varlen'


class LyricsEvent(MetaEventWithText):
    name = 'Lyrics'
    metacommand = 0x05
    length = 'varlen'


class MarkerEvent(MetaEventWithText):
    name = 'Marker'
    metacommand = 0x06
    length = 'varlen'


class CuePointEvent(MetaEventWithText):
    name = 'Cue Point'
    metacommand = 0x07
    length = 'varlen'


class ProgramNameEvent(MetaEventWithText):
    name = 'Program Name'
    metacommand = 0x08
    length = 'varlen'


class UnknownMetaEvent(MetaEvent):
    name = 'Unknown'
    # This class variable must be overriden by code calling the constructor,
    # which sets a local variable of the same name to shadow the class
    # variable.
    metacommand = None

    def __init__(self, **kw):
        super(MetaEvent, self).__init__(**kw)
        self.metacommand = kw['metacommand']

    def copy(self, **kw):
        kw['metacommand'] = self.metacommand
        return super(UnknownMetaEvent, self).copy(kw)


class ChannelPrefixEvent(MetaEvent):
    name = 'Channel Prefix'
    metacommand = 0x20
    length = 1


class PortEvent(MetaEvent):
    name = 'MIDI Port/Cable'
    metacommand = 0x21


class TrackLoopEvent(MetaEvent):
    name = 'Track Loop'
    metacommand = 0x2E


class EndOfTrackEvent(MetaEvent):
    name = 'End of Track'
    metacommand = 0x2F


class SetTempoEvent(MetaEvent):
    # __slots__ = ['bpm', 'mpqn']
    name = 'Set Tempo'
    metacommand = 0x51
    length = 3

    def set_bpm(self, bpm):
        self.mpqn = int(float(6e7) / bpm)

    def get_bpm(self):
        return float(6e7) / self.mpqn
    bpm = property(get_bpm, set_bpm)

    def get_mpqn(self):
        assert(len(self.data) == 3)
        vals = [self.data[x] << (16 - (8 * x)) for x in range(3)]
        return sum(vals)

    def set_mpqn(self, val):
        self.data = [(val >> (16 - (8 * x)) & 0xFF) for x in range(3)]
    mpqn = property(get_mpqn, set_mpqn)


class SmpteOffsetEvent(MetaEvent):
    name = 'SMPTE Offset'
    metacommand = 0x54


class TimeSignatureEvent(MetaEvent):
    # __slots__ = ['numerator', 'denominator', 'metronome', 'thirtyseconds']
    name = 'Time Signature'
    metacommand = 0x58
    length = 4

    def get_numerator(self):
        return self.data[0]

    def set_numerator(self, val):
        self.data[0] = val
    numerator = property(get_numerator, set_numerator)

    def get_denominator(self):
        return 2 ** self.data[1]

    def set_denominator(self, val):
        self.data[1] = int(math.log(val, 2))
    denominator = property(get_denominator, set_denominator)

    def get_metronome(self):
        return self.data[2]

    def set_metronome(self, val):
        self.data[2] = val
    metronome = property(get_metronome, set_metronome)

    def get_thirtyseconds(self):
        return self.data[3]

    def set_thirtyseconds(self, val):
        self.data[3] = val
    thirtyseconds = property(get_thirtyseconds, set_thirtyseconds)


class KeySignatureEvent(MetaEvent):
    # __slots__ = ['alternatives', 'minor']
    name = 'Key Signature'
    metacommand = 0x59
    length = 2

    def get_alternatives(self):
        d = self.data[0]
        return d - 256 if d > 127 else d

    def set_alternatives(self, val):
        self.data[0] = 256 + val if val < 0 else val
    alternatives = property(get_alternatives, set_alternatives)

    def get_minor(self):
        return self.data[1]

    def set_minor(self, val):
        self.data[1] = val
    minor = property(get_minor, set_minor)


class SequencerSpecificEvent(MetaEvent):
    name = 'Sequencer Specific'
    metacommand = 0x7F



##
## Constants
##

OCTAVE_MAX_VALUE = 12
OCTAVE_VALUES = list(range( OCTAVE_MAX_VALUE))

NOTE_NAMES = ['C','Cs','D','Ds','E','F','Fs','G','Gs','A','As','B']
WHITE_KEYS = [0, 2, 4, 5, 7, 9, 11]
BLACK_KEYS = [1, 3, 6, 8, 10]
NOTE_PER_OCTAVE = len( NOTE_NAMES )
NOTE_VALUES = list(range( OCTAVE_MAX_VALUE * NOTE_PER_OCTAVE))
NOTE_NAME_MAP_FLAT = {}
NOTE_VALUE_MAP_FLAT = []
NOTE_NAME_MAP_SHARP = {}
NOTE_VALUE_MAP_SHARP = []

for value in range( 128 ):
    noteidx = value % NOTE_PER_OCTAVE
    octidx = value / OCTAVE_MAX_VALUE
    name = NOTE_NAMES[noteidx]
    if len( name ) == 2:
        # sharp note
        flat = NOTE_NAMES[noteidx+1] + 'b'
        NOTE_NAME_MAP_FLAT['%s_%d' % (flat, octidx)] = value
        NOTE_NAME_MAP_SHARP['%s_%d' % (name, octidx)] = value
        NOTE_VALUE_MAP_FLAT.append( '%s_%d' % (flat, octidx) )
        NOTE_VALUE_MAP_SHARP.append( '%s_%d' % (name, octidx) )
        globals()['%s_%d' % (name[0] + 's', octidx)] = value
        globals()['%s_%d' % (flat, octidx)] = value
    else:
        NOTE_NAME_MAP_FLAT['%s_%d' % (name, octidx)] = value
        NOTE_NAME_MAP_SHARP['%s_%d' % (name, octidx)] = value
        NOTE_VALUE_MAP_FLAT.append( '%s_%d' % (name, octidx) )
        NOTE_VALUE_MAP_SHARP.append( '%s_%d' % (name, octidx) )
        globals()['%s_%d' % (name, octidx)] = value

BEATNAMES = ['whole', 'half', 'quarter', 'eighth', 'sixteenth', 'thiry-second', 'sixty-fourth']
BEATVALUES = [4, 2, 1, .5, .25, .125, .0625]
WHOLE = 0
HALF = 1
QUARTER = 2
EIGHTH = 3
SIXTEENTH = 4
THIRTYSECOND = 5
SIXTYFOURTH = 6

DEFAULT_MIDI_HEADER_SIZE = 14




import inspect
import math
import typing as tp
from collections import OrderedDict
from abc import abstractmethod, ABCMeta
from beatsearch.rhythm import MonophonicRhythm, PolyphonicRhythm, Unit, UnitType, parse_unit_argument
from beatsearch.feature_extraction import BinaryOnsetVector, IOIVector, \
    IOIDifferenceVector, OnsetPositionVector, ChronotonicChain
from beatsearch.utils import friendly_named_class, Quantizable


class DistanceMeasure(object):
    """Abstract base class for distance measures"""

    def get_distance(self, obj_a, obj_b):
        raise NotImplementedError


class MonophonicRhythmDistanceMeasure(DistanceMeasure, metaclass=ABCMeta):
    """Abstract base class for monophonic rhythm distance measures

    This is an abstract base class for monophonic rhythm distance measures. It measures the distance
    between two MonophonicRhythmImpl objects.
    """

    LENGTH_POLICIES = ["exact", "multiple", "fill"]
    DEFAULT_LENGTH_POLICY = LENGTH_POLICIES[0]

    class UnknownLengthPolicy(Exception):
        def __init__(self, given_length_policy):
            super(MonophonicRhythmDistanceMeasure.UnknownLengthPolicy, self).__init__(
                "Given %s, please choose between: %s" % (
                    given_length_policy, MonophonicRhythmDistanceMeasure.LENGTH_POLICIES))

    class LengthPolicyNotMet(Exception):
        MESSAGES = {
            'exact': "When length policy is set to \"exact\", both iterables should have the same number of elements",
            'multiple': "When length policy is set to \"multiple\", the length of the largest "
                        "iterable should be a multiple of all the other iterable lengths"
        }

        def __init__(self, length_policy):
            msg = self.MESSAGES[length_policy]
            super().__init__(msg)

    def __init__(self, unit: tp.Optional[UnitType] = None, length_policy=DEFAULT_LENGTH_POLICY):
        """
        Creates a new monophonic rhythm distance measure.

        :param unit: see unit property
        :param length_policy: see documentation on length_policy property
        """

        self._len_policy = ''   # type: str
        self._unit = None       # type: tp.Union[None, Unit]

        self.length_policy = length_policy  # calling setter
        self.set_unit(unit)

    @property
    def length_policy(self):
        """
        The length policy determines how permissive the monophonic rhythm similarity measure is with variable sized
        rhythm vector (N = onset count) or chain (N = duration) representations. Given two rhythm representations X and
        Y, the length policy should be one of:

            'exact': len(X) must equal len(Y)
            'multiple': len(X) must be a multiple of len(Y) or vice-versa
            'fill': len(X) and len(Y) must not be empty

        Implementations of MonophonicRhythmDistanceMeasure.get_distance will throw a ValueError if the representations
        of the given rhythm do not meet the requirements according to the length policy.
        """

        return self._len_policy

    @length_policy.setter
    def length_policy(self, length_policy):
        valid_policies = self.LENGTH_POLICIES
        if length_policy not in valid_policies:
            raise self.UnknownLengthPolicy(length_policy)
        self._len_policy = length_policy

    @parse_unit_argument
    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        """Sets the time unit

        This time unit is used internally for distance computation. When the unit is set to None, distance computation
        will be tick-based.

        :param unit: Unit enum object or None for tick-based distance computation
        :return: None
        """

        self._unit = unit

    def get_unit(self) -> tp.Union[Unit, None]:
        """Returns the time unit

        This time unit is used internally for distance computation. If the unit is None, distance computation is
        tick-based.

        :return: Unit enum object or None
        """

        return self._unit

    @property
    def unit(self) -> tp.Union[Unit, None]:
        """Unit used internally for distance computation."""
        return self.get_unit()

    @unit.setter
    def unit(self, unit):
        self.set_unit(unit)

    def get_distance(self, rhythm_a: MonophonicRhythm, rhythm_b: MonophonicRhythm) -> tp.Any:
        """
        Returns the distance between the given tracks. If no unit is set the computation will be tick-based and both
        rhythms must have the same tick resolution.

        :param rhythm_a: monophonic rhythm to compare to monophonic rhythm b
        :param rhythm_b: monophonic rhythm to compare to monophonic rhythm a
        :return: distance between the given monophonic rhythms
        :raises ValueError: if no unit is set and given rhythms don't have the same tick resolution
        """

        tick_based = self.unit is None
        res_a, res_b = rhythm_a.get_resolution(), rhythm_b.get_resolution()

        if tick_based and res_a != res_b:
            raise ValueError("Can't do a tick-based distance computation for rhythms "
                             "with different tick resolutions (%i vs %i)" % (res_a, res_b))

        rhythms = [rhythm_a, rhythm_b]
        iterables = [self.__get_iterable__(t) for t in rhythms]
        cookies = [self.__get_cookie__(t) for t in rhythms]

        if not self.__class__.check_if_iterables_meet_len_policy(self.length_policy, *iterables):
            raise self.LengthPolicyNotMet(self.length_policy)

        return self.__compute_distance__(max(len(i) for i in iterables), *(iterables + cookies))

    @abstractmethod
    def __get_iterable__(self, rhythm: MonophonicRhythm):
        """
        Should prepare and return the rhythm representation on which the similarity measure is based. The returned
        vector's will be length policy checked. The returned vector should be in the current unit.

        :param rhythm: the monophonic rhythm
        :return: desired rhythm representation to use in __compute_distance__
        """

        raise NotImplementedError

    # noinspection PyMethodMayBeStatic
    def __get_cookie__(self, rhythm: MonophonicRhythm):
        """
        The result of this method will be passed to __compute_distance__, both for rhythm a and rhythm b. By default,
        the cookie is the rhythm itself.

        :param rhythm: the monophonic rhythm
        :return: cookie to use in __compute_distance__
        """

        return rhythm

    @staticmethod  # TODO enforce python 3.3
    @abstractmethod
    def __compute_distance__(max_len, iterable_a, iterable_b, cookie_a, cookie_b):
        """
        The result of this method is returned by get_distance. If that method is given two rhythms a and b, this method
        is given both the iterables of a and b and the cookies of a and b, returned by respectively __get_iterable__
        and __get_cookie__.

        :param max_len: max(len(iterable_a), len(iterable_b))
        :param iterable_a: the result of __get_iterable__, given rhythm a
        :param iterable_b: the result of __get_iterable__, given rhythm b
        :param cookie_a: the result of __get_cookie__, given rhythm a
        :param cookie_b: the result of __get_cookie__, given rhythm b
        :return: the distance between rhythm a and b
        """

        raise NotImplementedError

    @classmethod
    def check_if_iterables_meet_len_policy(cls, length_policy, iterable_a, iterable_b):
        """
        Checks whether the given iterables meet a certain duration policy. Duration policies are:
            'exact'     - met when all iterables have the exact same length and are not empty
            'multiple'  - met when the length of the largest iterable is a multiple of all the other iterable lengths
            'fill'      - met when any of the chains is empty

        :param length_policy: one of {'exact', 'multiple' or 'fill'}
        :param iterable_a: iterable
        :param iterable_b: iterable
        :return: True if the two iterables meet the length policy, False if not
        :raises ValueError if given unknown length policy
        """

        l = [len(iterable_a), len(iterable_b)]

        if length_policy == "exact":
            return l[0] == l[1]
        elif length_policy == "multiple":
            return all(x % l[0] == 0 or l[0] % x == 0 for x in l)
        elif length_policy == "fill":
            return l[0] > 0 and l[1] > 0

        raise cls.UnknownLengthPolicy(length_policy)

    __measures_by_friendly_name__ = {}  # monophonic rhythm distance implementations by __friendly_name__
    __measures_by_class_name__ = {}  # monophonic rhythm distance implementations by __name__

    @classmethod
    def get_measures(cls, friendly_name=True):
        """
        Returns an ordered dictionary containing implementations of MonophonicRhythmDistanceMeasure by name.

        :param friendly_name: when True, the name will be retrieved with __friendly_name__ instead of __name__
        :return: an ordered dictionary containing all subclasses of MonophonicRhythmDistanceMeasure by name
        """

        if len(MonophonicRhythmDistanceMeasure.__measures_by_friendly_name__) != \
                len(MonophonicRhythmDistanceMeasure.__subclasses__()):
            by_friendly_name = OrderedDict()
            by_class_name = OrderedDict()
            for dm in cls.__subclasses__():  # Distance Measure
                name = dm.__name__
                if friendly_name:
                    try:
                        # noinspection PyUnresolvedReferences
                        name = dm.__friendly_name__
                    except AttributeError:
                        pass
                by_friendly_name[name] = dm
                by_class_name[dm.__name__] = dm
            cls.__measures_by_friendly_name__ = by_friendly_name
            cls.__measures_by_class_name__ = by_class_name

        return MonophonicRhythmDistanceMeasure.__measures_by_friendly_name__

    @classmethod
    def get_measure_names(cls):
        measures = cls.get_measures()
        return tuple(measures.keys())

    @classmethod
    def get_measure_by_name(cls, measure_name):
        """
        Returns a subclass given its name. This can either be its friendly name or its class name.

        :param measure_name: either its friendly name or its class name
        :return: the monophonic rhythm distance measure
        """

        measures = cls.get_measures()

        try:
            return measures[measure_name]
        except KeyError:
            by_class_name = cls.__measures_by_class_name__
            try:
                return by_class_name[measure_name]
            except KeyError:
                raise ValueError("No measure with class or friendly name: '%s'" % measure_name)


@friendly_named_class("Hamming distance")
class HammingDistanceMeasure(MonophonicRhythmDistanceMeasure):
    """
    The hamming distance is based on the binary chains of the rhythms. The hamming distance is the sum of indexes
    where the binary rhythm chains do not match. The hamming distance is always an integer.
    """

    def __init__(self, unit: tp.Optional[UnitType] = Unit.EIGHTH, length_policy="multiple"):
        self._binary_onset_vector = BinaryOnsetVector(unit)
        super().__init__(unit, length_policy)

    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        super().set_unit(unit)
        self._binary_onset_vector.set_unit(unit)

    def __get_iterable__(self, rhythm: MonophonicRhythm):
        assert self._binary_onset_vector.unit == self.unit
        return self._binary_onset_vector.process(rhythm)

    @staticmethod
    def __compute_distance__(n, cx, cy, *cookies):
        hamming_distance, i = 0, 0
        while i < n:
            x = cx[i % len(cx)]
            y = cy[i % len(cy)]
            hamming_distance += x != y
            i += 1
        return hamming_distance


@friendly_named_class("Euclidean interval vector distance")
class EuclideanIntervalVectorDistanceMeasure(MonophonicRhythmDistanceMeasure, Quantizable):
    """
    The euclidean interval vector distance is the euclidean distance between the inter-onset vectors of the rhythms.
    """

    def __init__(self, unit: tp.Optional[UnitType] = None, length_policy="exact", quantize=False):
        self._ioi_vector = IOIVector(unit, mode=IOIVector.Mode.POST_NOTE, quantize_enabled=quantize)
        super().__init__(unit, length_policy)  # super constructor must be called after self._ioi_vector definition
        self.quantize_enabled = quantize

    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        super().set_unit(unit)
        self._ioi_vector.set_unit(unit)

    def set_quantize_enabled(self, quantize_enabled):
        self._ioi_vector.set_quantize_enabled(quantize_enabled)

    def is_quantize_enabled(self) -> bool:
        return self._ioi_vector.is_quantize_enabled()

    def __get_iterable__(self, rhythm: MonophonicRhythm):
        assert self._ioi_vector.unit == self.unit
        return self._ioi_vector.process(rhythm)

    @staticmethod
    def __compute_distance__(n, vx, vy, *cookies):
        sum_squared_dt, i = 0, 0
        while i < n:
            dt = vx[i % len(vx)] - vy[i % len(vy)]
            sum_squared_dt += dt * dt
            i += 1
        return math.sqrt(sum_squared_dt)


@friendly_named_class("Interval difference vector distance")
class IntervalDifferenceVectorDistanceMeasure(MonophonicRhythmDistanceMeasure, Quantizable):
    """
    The interval difference vector distance is based on the interval difference vectors of the rhythms.
    """

    def __init__(self, unit: tp.Optional[UnitType] = None, length_policy="fill", quantize=False, cyclic=True):
        self._ioi_diff_vector = IOIDifferenceVector(unit, cyclic)
        super().__init__(unit, length_policy)
        self.quantize_enabled = quantize
        self.cyclic = cyclic  # call setter

    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        super().set_unit(unit)
        self._ioi_diff_vector.set_unit(unit)

    @property
    def cyclic(self):
        return self._ioi_diff_vector.cyclic

    @cyclic.setter
    def cyclic(self, cyclic):
        self._ioi_diff_vector.cyclic = cyclic

    def set_quantize_enabled(self, quantize_enabled: bool):
        self._ioi_diff_vector.set_quantize_enabled(quantize_enabled)

    def is_quantize_enabled(self) -> bool:
        return self._ioi_diff_vector.is_quantize_enabled()

    def __get_iterable__(self, rhythm: MonophonicRhythm):
        assert self._ioi_diff_vector.unit == self.unit
        return self._ioi_diff_vector.process(rhythm)

    @staticmethod
    def __compute_distance__(n, vx, vy, *cookies):
        summed_fractions, i = 0, 0
        while i < n:
            x = float(vx[i % len(vx)])
            y = float(vy[i % len(vy)])
            numerator, denominator = (x, y) if x > y else (y, x)
            try:
                summed_fractions += numerator / denominator
            except ZeroDivisionError:
                return float("inf")
            i += 1
        return summed_fractions - n


@friendly_named_class("Swap distance")
class SwapDistanceMeasure(MonophonicRhythmDistanceMeasure, Quantizable):
    """
    The swap distance is the minimal number of swap operations required to transform one rhythm to another. A swap is an
    interchange of a one and a zero that are adjacent to each other in the binary representations of the rhythms.

    Although the concept of the swap distance is based on the rhythm's binary chain, this implementation uses the
    absolute onset times of the onsets. This makes it possible to work with floating point swap operations (0.x swap
    operation). Enable this by setting quantize to True in the constructor.
    """

    def __init__(self, unit: tp.Optional[UnitType] = Unit.EIGHTH, length_policy="multiple", quantize=False):
        self._onset_position_vector = OnsetPositionVector(unit, quantize)
        super().__init__(unit, length_policy)  # super constructor must be called after _onset_position_vector attr def
        self.quantize_enabled = quantize

    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        super().set_unit(unit)
        self._onset_position_vector.set_unit(unit)

    def set_quantize_enabled(self, quantize_enabled: bool):
        self._onset_position_vector.set_quantize_enabled(quantize_enabled)

    def is_quantize_enabled(self) -> bool:
        return self._onset_position_vector.is_quantize_enabled()

    def __get_iterable__(self, rhythm: MonophonicRhythm):
        assert self._onset_position_vector.unit == self.unit
        return self._onset_position_vector.process(rhythm)

    def __get_cookie__(self, rhythm: MonophonicRhythm):
        return rhythm.get_duration(self.unit, ceil=True)

    @staticmethod
    def __compute_distance__(n, vx, vy, dur_x, dur_y):
        swap_distance, i, x_offset, y_offset = 0, 0, 0, 0
        while i < n:
            x_offset = i // len(vx) * dur_x
            y_offset = i // len(vy) * dur_y
            x = vx[i % len(vx)] + x_offset
            y = vy[i % len(vy)] + y_offset
            swap_distance += abs(x - y)
            i += 1
        return swap_distance


@friendly_named_class("Chronotonic distance")
class ChronotonicDistanceMeasure(MonophonicRhythmDistanceMeasure):
    """
    The chronotonic distance is the area difference (aka measure K) of the rhythm's chronotonic chains.
    """

    def __init__(self, unit: tp.Optional[UnitType] = Unit.EIGHTH, length_policy="multiple"):
        self._chronotonic_vector = ChronotonicChain(unit)
        super().__init__(unit, length_policy)  # super constructor must be called after _chronotonic_vector attr def

    def set_unit(self, unit: tp.Optional[UnitType]) -> None:
        super().set_unit(unit)
        self._chronotonic_vector.set_unit(unit)

    def __get_iterable__(self, rhythm: MonophonicRhythm):
        assert self._chronotonic_vector.unit == self.unit
        return self._chronotonic_vector.process(rhythm)

    @staticmethod
    def __compute_distance__(n, cx, cy, *args):
        chronotonic_distance, i = 0, 0
        while i < n:
            x = cx[i % len(cx)]
            y = cy[i % len(cy)]
            # assuming that each pulse is a unit
            chronotonic_distance += abs(x - y)
            i += 1
        return chronotonic_distance


TRACK_WILDCARDS = ["*", "a*", "b*"]  # NOTE: Don't change the wildcard order or the code will break


def rhythm_pair_track_iterator(rhythm_a: PolyphonicRhythm,
                               rhythm_b: PolyphonicRhythm,
                               tracks: tp.Union[str, tp.Iterable[tp.Any]]):
    """
    Returns an iterator over the tracks of the given rhythms. Each iteration yields a track name and a pair of tracks
    with that track name; rhythm_a.get_track(name) and rhythm_b.get_track(name). This is yielded as a tuple:
        (track_name, (track_a, track_b))

    The tracks to iterate over can be specified with the 'tracks' argument:

        [iterable] - Iterate over tracks with the names in the given iterable
        '*'        - Iterate over all track names. If the rhythms contain tracks with the same name, this will just
                     result in one iteration. E.g. if both rhythms contain a track named 'foo', this will result in one
                     iteration yielding ('foo', (track_a, track_b))).
        'a*'      - Iterate over track names in rhythm a
        'b*'      - Iterate over track names in rhythm b

    :param rhythm_a: the first rhythm
    :param rhythm_b: the second rhythm
    :param tracks: either an iterable containing the track names, or one of the wildcards ['*', 'a*' or 'b*']
    :return: an iterator over the tracks of the given rhythms
    """

    try:
        wildcard_index = TRACK_WILDCARDS.index(tracks)
    except ValueError:
        wildcard_index = -1
        try:
            tracks = iter(tracks)
        except TypeError:
            raise ValueError("Excepted an iterable or "
                             "one of %s but got %s" % (TRACK_WILDCARDS, tracks))

    it_a, it_b = rhythm_a.get_track_iterator(), rhythm_b.get_track_iterator()

    if wildcard_index == -1:  # if given specific tracks
        for track_name in tracks:
            track_a = rhythm_a.get_track_by_name(track_name)
            track_b = rhythm_b.get_track_by_name(track_name)
            yield (track_name, (track_a, track_b))

    elif wildcard_index == 0:  # if '*' wildcard
        names = set()
        for track_a in it_a:
            name = track_a.name
            names.add(name)
            track_b = rhythm_b.get_track_by_name(name)
            names.add(name)
            yield (name, (track_a, track_b))
        for track_b in it_b:
            name = track_b.name
            if name in names:
                continue
            track_a = rhythm_a.get_track_by_name(name)
            yield (name, (track_a, track_b))

    elif wildcard_index == 1:  # if wildcard 'a*'
        for track_a in it_a:
            name = track_a.name
            track_b = rhythm_b.get_track_by_name(name)
            yield (name, (track_a, track_b))

    elif wildcard_index == 2:  # if wildcard 'b*'
        for track_b in it_b:
            name = track_b.name
            track_a = rhythm_a.get_track_by_name(name)
            yield (name, (track_a, track_b))

    else:
        assert False


class PolyphonicRhythmDistanceMeasure(DistanceMeasure, metaclass=ABCMeta):
    """Abstract base class for polyphonic rhythm distance measures

    This is an abstract base class for polyphonic rhythm distance measures. It measures the distance
    between two Rhythm objects.
    """

    @abstractmethod
    def get_distance(self, rhythm_a: PolyphonicRhythm, rhythm_b: PolyphonicRhythm) -> tp.Union[float, int]:
        raise NotImplementedError


class SummedMonophonicRhythmDistance(PolyphonicRhythmDistanceMeasure):
    def __init__(self, track_distance_measure=HammingDistanceMeasure, tracks="a*", normalize=True):
        # type: (tp.Union[MonophonicRhythmDistanceMeasure, type], tp.Union[str, tp.Iterable[tp.Any]]) -> None
        self._tracks = []
        self.tracks = tracks
        self._track_distance_measure = None
        self.monophonic_measure = track_distance_measure
        self.normalize = normalize

    @property
    def monophonic_measure(self):  # type: () -> MonophonicRhythmDistanceMeasure
        """
        The distance measure used to compute the distance between the rhythm tracks; an instance of
        MonophonicRhythmDistanceMeasure.
        """

        return self._track_distance_measure

    @monophonic_measure.setter
    def monophonic_measure(
            self,
            track_distance_measure: tp.Union[MonophonicRhythmDistanceMeasure, tp.Type[MonophonicRhythmDistanceMeasure]]
    ):
        """
        Setter for the track distance measure. This should be either a MonophonicRhythmDistanceMeasure subclass or
        instance. When given a class, the measure will be initialized with no arguments.
        """

        if inspect.isclass(track_distance_measure) and \
                issubclass(track_distance_measure, MonophonicRhythmDistanceMeasure):
            track_distance_measure = track_distance_measure()
        elif not isinstance(track_distance_measure, MonophonicRhythmDistanceMeasure):
            raise ValueError("Expected a MonophonicRhythmDistanceMeasure subclass or "
                             "instance, but got '%s'" % track_distance_measure)
        self._track_distance_measure = track_distance_measure

    @property
    def tracks(self):
        """
        The tracks to iterate over when computing the distance. See rhythm_pair_track_iterator.
        """

        return self._tracks

    @tracks.setter
    def tracks(self, tracks):
        self._tracks = tracks

    def get_distance(self, rhythm_a, rhythm_b):
        """
        Returns the average track distance from the tracks of rhythm a and b. When the track distance can't be computed,
        duration of the longest rhythm is used as a distance. This can either happen when the a certain track doesn't
        exist in the two rhythms (e.g. 'snare' track in rhythm_a but not in rhythm_b) or when the track distance measure
        raises an exception when computing the distance.

        See RhythmSimilarityMeasure.tracks to specify which tracks should measured.

        :param rhythm_a: the first rhythm
        :param rhythm_b: the second rhythm
        :return: the average track distance of the two rhythms
        """

        measure = self.monophonic_measure
        unit = measure.unit  # type: tp.Union[Unit, None]
        duration = max(rhythm_a.get_duration(unit), rhythm_b.get_duration(unit))
        n_tracks, total_distance = 0, 0

        for name, tracks in rhythm_pair_track_iterator(rhythm_a, rhythm_b, self.tracks):
            distance = duration
            if None not in tracks:
                try:
                    distance = measure.get_distance(tracks[0], tracks[1])
                except MonophonicRhythmDistanceMeasure.LengthPolicyNotMet:
                    pass
            total_distance += distance
            n_tracks += 1

        average_distance = float(total_distance) / n_tracks
        return average_distance / duration if self.normalize else average_distance


__all__ = [
    # Distance measure abstract base classes
    'DistanceMeasure', 'MonophonicRhythmDistanceMeasure', 'PolyphonicRhythmDistanceMeasure',

    # Monophonic rhythm distance measure implementations
    'HammingDistanceMeasure', 'EuclideanIntervalVectorDistanceMeasure',
    'IntervalDifferenceVectorDistanceMeasure', 'SwapDistanceMeasure', 'ChronotonicDistanceMeasure',

    # Polyphonic rhythm distance measure implementations
    'SummedMonophonicRhythmDistance'
]
